def atomwise_tokenizer(smi, exclusive_tokens=None):
    """
    Tokenize a SMILES molecule at atom-level and return tokens with their token IDs.
    - 'Br' and 'Cl' are two-character tokens.
    - Symbols with brackets are considered as tokens.
    - If `exclusive_tokens` is provided, symbols with brackets not in `exclusive_tokens` will be replaced by '[UNK]'.

    Parameters:
    - smi (str): SMILES string to tokenize.
    - exclusive_tokens (list of str, optional): Specific symbols with brackets to keep.

    Returns:
    - tuple: (tokens, token_ids), where tokens is a list of atom-level tokens and token_ids is a list of corresponding token IDs.
    """
    import re
    
    # Regular expression pattern to match atom-level tokens in a SMILES string
    pattern = "(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\\\|\/|:|~|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])"
    regex = re.compile(pattern)
    
    # Find all tokens in the SMILES string based on the regex pattern
    tokens = [token for token in regex.findall(smi)]

    # Handle exclusive tokens, replacing non-exclusive bracketed tokens with '[UNK]'
    if exclusive_tokens:
        tokens = [tok if tok in exclusive_tokens or not tok.startswith('[') else '[UNK]' for tok in tokens]

    # Generate token IDs based on the order of unique token appearance
    token_to_id = {}  # Dictionary to map tokens to their IDs
    token_ids = []  # List to store token IDs
    for token in tokens:
        if token not in token_to_id:
            # Assign a new ID based on the current size of the dictionary
            token_to_id[token] = len(token_to_id)
        token_ids.append(token_to_id[token])

    return tokens, token_ids  # Return the tokens and their corresponding IDs