import argparse
import re
from typing import List

from datasets import load_dataset
from tokenizers import Tokenizer, Regex, NormalizedString, PreTokenizedString
from tokenizers.models import WordLevel, BPE
from tokenizers.pre_tokenizers import Split, PreTokenizer, Whitespace
from tokenizers.processors import TemplateProcessing
from tokenizers.trainers import WordLevelTrainer, BpeTrainer

from SmilesPE.learner import *
from SmilesPE.tokenizer import *

import codecs

# Token for accessing datasets (placeholder)
Token = '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'

def atomwise_tokenizer(smi: str, exclusive_tokens=None) -> List[str]:
    """Tokenize a SMILES molecule at atom-level:
        (1) 'Br' and 'Cl' are two-character tokens
        (2) Symbols with brackets are considered as tokens

    :param smi: A SMILES string
    :param exclusive_tokens: A list of specific symbols with brackets you want to keep. e.g., ['[C@@H]', '[nH]'], defaults to None
    :return: A list of atom-wise tokens
    """
    pattern = "(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\\\|\/|:|~|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])"  # type: ignore
    regex = re.compile(pattern)
    tokens = [token for token in regex.findall(smi)]

    if exclusive_tokens:
        for i, tok in enumerate(tokens):
            if tok.startswith("["):
                if tok not in exclusive_tokens:
                    tokens[i] = "<unk>"  # Replace non-exclusive tokens with <unk>
    return tokens

def tokens_to_mer(toks: List[str]) -> str:
    """Convert a list of tokens to a k-mer token.

    :param toks: A list of tokens
    :return: A k-mer token
    """
    return "".join(toks)

def kmer_tokenizer(
    smiles: str,
    ngram: int = 4,
    stride: int = 1,
    remove_last: bool = False,
    exclusive_tokens=None,
) -> List[str]:
    """Tokenize a SMILES molecule at k-mer level.

    :param smiles: A SMILES string
    :param ngram: k-mer size, defaults to 4
    :param stride: k-mer stride, defaults to 1
    :param remove_last: Whether to remove the last token, defaults to False
    :param exclusive_tokens: A list of specific symbols with brackets you want to keep. e.g., ['[C@@H]', '[nH]'], defaults to None
    :return: A list of k-mer tokens
    """
    # Tokenize at atom level first
    units = atomwise_tokenizer(smiles, exclusive_tokens=exclusive_tokens)
    
    # Generate k-mer tokens
    if ngram == 1:
        tokens = units
    else:
        tokens = [
            tokens_to_mer(units[i: i + ngram])
            for i in range(0, len(units), stride)
            if len(units[i: i + ngram]) == ngram
        ]

    # Optionally remove the last token if its length is less than ngram
    if remove_last and len(tokens[-1]) < ngram:
        tokens = tokens[:-1]
    return tokens

def load_bpe_tokenizer(filepath: str) -> Tokenizer:
    """
    Load a BPE tokenizer from a file.

    :param filepath: Path to the saved tokenizer.
    :return: Loaded Tokenizer object.
    """
    tokenizer = Tokenizer.from_file(filepath)
    return tokenizer

def load_smilespe_tokenizer(filepath: str):
    """
    Load a SmilesPE tokenizer from a file.

    :param filepath: Path to the saved tokenizer.
    :return: Loaded Tokenizer object.
    """
    spe_vocab = codecs.open(filepath)
    return SPE_Tokenizer(spe_vocab)

def bpe_tokenizer(smiles: str, tokenizer: Tokenizer) -> List[str]:
    """
    Tokenize a SMILES string using a BPE tokenizer.

    :param smiles: A SMILES string to tokenize.
    :param tokenizer: A loaded BPE tokenizer.
    :return: A list of tokens.
    """
    output = tokenizer.encode(smiles)
    tokens = output.tokens
    return tokens

def smilespe_tokenizer(smiles: str, tokenizer: Tokenizer) -> List[str]:
    """
    Tokenize a SMILES string using a SmilesPE tokenizer.

    :param smiles: A SMILES string to tokenize.
    :param tokenizer: A loaded SmilesPE tokenizer.
    :return: A list of tokens.
    """
    tokens = tokenizer.tokenize(smiles)
    return tokens

class KmerPreTokenizer:
    """Pre-tokenizer for k-mer tokenization."""
    def __init__(
        self,
        ngram: int = 4,
        stride: int = 1,
        remove_last: bool = False,
        exclusive_tokens=None,
    ):
        self.ngram = ngram
        self.stride = stride
        self.remove_last = remove_last
        self.exclusive_tokens = exclusive_tokens

    def kmer_split(
        self, i: int, normalized_string: NormalizedString
    ) -> List[NormalizedString]:
        """Split a normalized string into k-mers."""
        splits = []
        for token in kmer_tokenizer(
            str(normalized_string),
            ngram=self.ngram,
            stride=self.stride,
            remove_last=self.remove_last,
            exclusive_tokens=self.exclusive_tokens,
        ):
            splits.append(NormalizedString(token))  # type: ignore
        return splits

    def pre_tokenize(self, pretok: PreTokenizedString):
        """Pre-tokenize a string into k-mers."""
        pretok.split(self.kmer_split)

def train_tokenizer(
    dataset_name: str = "PubChem",
    mol_type: str = "SMILES",
    tokenizer_method: str = "atomwise",
    ngram: int = 4,
    stride: int = 1,
    remove_last: bool = False,
    exclusive_tokens=None,
    max_vocab_size: int = 30000,
) -> None:
    """Train a tokenizer on a dataset of SMILES strings.

    :param dataset_name: Name of the dataset, defaults to "PubChem"
    :param mol_type: Molecule type, defaults to "SMILES"
    :param tokenizer_method: Tokenization method, defaults to "atomwise"
    :param ngram: k-mer size for k-mer tokenization, defaults to 4
    :param stride: k-mer stride for k-mer tokenization, defaults to 1
    :param remove_last: Whether to remove the last token, defaults to False
    :param exclusive_tokens: List of specific tokens to keep, defaults to None
    :param max_vocab_size: Maximum vocabulary size, defaults to 30000
    """
    # Load dataset
    data = load_dataset(f"MolGen/{dataset_name}", token=Token)
    smiles = data["train"][mol_type]  # type: ignore
    
    if tokenizer_method == 'smilespe':
        # Train a SmilesPE tokenizer
        outfile = codecs.open(f"{dataset_name}_{mol_type}_{tokenizer_method}_{max_vocab_size}.json", 'w')
        learn_SPE(smiles, outfile, num_symbols=30000, min_frequency=2000, augmentation=1, verbose=True, total_symbols=True)
        return

    # Initialize the tokenizer and trainer based on the method
    if tokenizer_method == "atomwise":
        tokenizer = Tokenizer(WordLevel(unk_token="<unk>"))  # type: ignore
        tokenizer.pre_tokenizer = Split(
            Regex(
                r"(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\\\|\/|:|~|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])"
            ),
            behavior="isolated",
        )  # type: ignore
        trainer = WordLevelTrainer(
            vocab_size=max_vocab_size,
            special_tokens=["<unk>", "<eos>", "<bos>", "<pad>"],
        )  # type: ignore
    elif tokenizer_method == "kmer":
        tokenizer = Tokenizer(WordLevel(unk_token="<unk>"))  # type: ignore
        tokenizer.pre_tokenizer = PreTokenizer.custom(
            KmerPreTokenizer(
                ngram=ngram,
                stride=stride,
                remove_last=remove_last,
                exclusive_tokens=exclusive_tokens,
            )
        )  # type: ignore
        trainer = WordLevelTrainer(
            vocab_size=max_vocab_size,
            special_tokens=["<unk>", "<eos>", "<bos>", "<pad>"],
        )  # type: ignore
    elif tokenizer_method == "bpe":
        tokenizer = Tokenizer(BPE(unk_token="<unk>"))
        trainer = BpeTrainer(
            vocab_size=max_vocab_size,
            special_tokens=["<unk>", "<bos>", "<eos>", "<pad>"],
        )  # type: ignore
    else:
        raise ValueError("Invalid tokenizer method")
        # Train the tokenizer on the SMILES data
tokenizer.train_from_iterator(smiles, trainer=trainer)

# Add post processing to the tokenizer
tokenizer.post_processor = TemplateProcessing(
    single="<bos> $A <eos>",
    special_tokens=[
        ("<bos>", 1),
        ("<eos>", 2),
    ],
)  # type: ignore

# Save the trained tokenizer
if tokenizer_method == "kmer":
    tokenizer.pre_tokenizer = Whitespace()  # type: ignore
    tokenizer.save(f"{dataset_name}_{mol_type}_{tokenizer_method}_{max_vocab_size}.json")
    
if name == "main":
    # Argument parser for command line options
    parser = argparse.ArgumentParser()
    parser.add_argument("-dataset_name", type=str, default="ZINC_250k")
    parser.add_argument("-mol_type", type=str, default="SMILES")
    parser.add_argument("-tokenizer_method", type=str, default="bpe")
    parser.add_argument("-max_vocab_size", type=int, default=30000)
    args = parser.parse_args()

    # Train the tokenizer with the provided arguments
    train_tokenizer(
        args.dataset_name, args.mol_type, args.tokenizer_method, args.max_vocab_size
    )

    # Example code to test the trained tokenizer (commented out)
    # loaded_bpe_tokenizer = load_bpe_tokenizer("ZINC_250k_SMILES_bpe_30000.json")
    # tokens = bpe_tokenizer("Brc1cccc(Nc2ncnc3ccncc23)c1NCCN1CCOCC1", loaded_bpe_tokenizer)
    # print(tokens)

    # loaded_smilespe_tokenizer = load_smilespe_tokenizer("ZINC_250k_SMILES_smilespe_30000.json")
    # tokens = smilespe_tokenizer("Brc1cccc(Nc2ncnc3ccncc23)c1NCCN1CCOCC1", loaded_smilespe_tokenizer)
    # print(tokens)